﻿#include <iostream>
#include <string>
#include <vector>
#include <memory>
#include <cstdint>
#include <cmath>
#include <sstream>
#include <exception>

using std::size_t;

class Writable {
public:
    virtual ~Writable() = default;
    virtual void write(std::ostream& out) const = 0;
    virtual Writable* clone() const = 0;
};

class String : public Writable {
public:
    String(const std::string& str): m_str(str) {}
    virtual ~String() override = default;
    virtual void write(std::ostream& out) const override { out << m_str; }
    virtual Writable* clone() const override {
        return new String(*this);
    };
private:
    std::string m_str;
};

class IDontLikeThatNumberException : public std::exception {
public:
    IDontLikeThatNumberException(double n): number(n) {
        std::ostringstream out;
        out << "I don't like number " << n << ", because cos(" << n << ") = " << std::cos(n) << "!";
        reason = out.str();
    }

    virtual const char* what() const noexcept override {
        return reason.c_str();
    }

    double number;
    std::string reason;
};


class Number : public Writable {
public:
    Number(double num): m_num(num) {
        if (std::cos(num) < 0) {
            throw IDontLikeThatNumberException(num);
        }
    }
    virtual ~Number() override = default;
    virtual void write(std::ostream& out) const override { out << m_num; }
    virtual Writable* clone() const override {
        return new Number(*this);
    }
private:
    double m_num;
};

class Group : public Writable {
public:
    Group(const std::string& prefix, const std::string& separator, const std::string& suffix):
        m_prefix(prefix),
        m_separator(separator),
        m_suffix(suffix) {}

    virtual ~Group() override = default;

    void push_back(std::unique_ptr<Writable> child) {
        m_children.push_back(std::move(child));
    }

    virtual void write(std::ostream& out) const override {
        out << m_prefix;
        bool first = true;
        for (auto& child : m_children) {
            if (!first) {
                out << m_separator;
            }
            first = false;
            child->write(out);
        }
        out << m_suffix;
    }

    void swap(Group& rhs) {
        m_children.swap(rhs.m_children);
        m_prefix.swap(rhs.m_prefix);
        m_separator.swap(rhs.m_separator);
        m_suffix.swap(rhs.m_suffix);
    }

    Group(Group&& rhs) {
        swap(rhs);
    }

    Group& operator=(Group&& rhs) {
        swap(rhs);
        return *this;
    }

    Group(const Group& rhs):
        m_children(),
        m_prefix(rhs.m_prefix),
        m_separator(rhs.m_separator),
        m_suffix(rhs.m_suffix) {

        for (const std::unique_ptr<Writable>& child : rhs.m_children) {
            m_children.push_back(std::unique_ptr<Writable>(child->clone()));
        }
    }

    Group& operator=(const Group& rhs) {
        auto tmp(rhs);
        swap(tmp);
        return *this;
    }

    virtual Writable* clone() const override {
        return new Group(*this);
    }

private:
    std::vector<std::unique_ptr<Writable>> m_children;
    std::string m_prefix;
    std::string m_separator;
    std::string m_suffix;
};

std::ostream& operator<<(std::ostream& out, const Writable& writable) {
    writable.write(out);
    return out;
}

int main() {
    try {
        Group g("[", " | ", "]");
        g.push_back(std::make_unique<Number>(123));
        g.push_back(std::make_unique<String>("abc"));

        Group g2("(", ", ", ")");
        g2.push_back(std::make_unique<Number>(456));
        g2.push_back(std::make_unique<String>("def"));
        g.push_back(std::make_unique<Group>(g2));

        g2.push_back(std::make_unique<String>("ghi"));

        std::cout << g << '\n';  // [123 | abc | (456, def)]
        std::cout << g2 << '\n'; // (456, def, ghi)
    } catch (const IDontLikeThatNumberException& e) {
        std::cerr << "IDontLikeThatNumberException was thrown:\n";
        std::cerr << e.what() << '\n';
    } catch (...) {
        std::cerr << "A totally unknown exception was thrown!\n";
    }
}